#[derive(Debug)]
pub struct AtemPacket<'packet_buffer> {
    flags: u8,
    session_id: u16,
    remote_packet_id: u16,
    retransmit_requested_from_packet_id: Option<u16>,
    ack_reply: Option<u16>,
    body: Option<&'packet_buffer [u8]>,
}

pub enum AtemPacketErr {
    TooShort(String),
    LengthDiffers(String),
}

#[derive(PartialEq)]
pub enum PacketFlag {
    AckRequest,
    NewSessionId,
    IsRetransmit,
    RetransmitRequest,
    AckReply,
}

impl From<PacketFlag> for u8 {
    fn from(flag: PacketFlag) -> Self {
        match flag {
            PacketFlag::AckRequest => 0x01,
            PacketFlag::NewSessionId => 0x02,
            PacketFlag::IsRetransmit => 0x04,
            PacketFlag::RetransmitRequest => 0x08,
            PacketFlag::AckReply => 0x10,
        }
    }
}

impl<'packet_buffer> AtemPacket<'packet_buffer> {
    pub fn session_id(&self) -> u16 {
        self.session_id
    }

    pub fn remote_packet_id(&self) -> u16 {
        self.remote_packet_id
    }

    pub fn body(&self) -> Option<&[u8]> {
        self.body
    }

    pub fn retransmit_request(&self) -> Option<u16> {
        self.retransmit_requested_from_packet_id
    }

    pub fn ack_reply(&self) -> Option<u16> {
        self.ack_reply
    }

    pub fn has_flag(&self, flag: PacketFlag) -> bool {
        self.flags & u8::from(flag) > 0
    }
}

impl<'packet_buffer> TryFrom<&'packet_buffer [u8]> for AtemPacket<'packet_buffer> {
    type Error = AtemPacketErr;

    fn try_from(buffer: &'packet_buffer [u8]) -> Result<Self, Self::Error> {
        if buffer.len() < 12 {
            return Err(AtemPacketErr::TooShort(format!(
                "Invalid packet from ATEM {:x?}",
                buffer
            )));
        }

        let length = u16::from_be_bytes([buffer[0], buffer[1]]) & 0x07ff;
        if length as usize != buffer.len() {
            return Err(AtemPacketErr::LengthDiffers(format!(
                "Length of message differs, expected {} got {}",
                length,
                buffer.len()
            )));
        }

        let flags = buffer[0] >> 3;
        let session_id = u16::from_be_bytes([buffer[2], buffer[3]]);
        let remote_packet_id = u16::from_be_bytes([buffer[10], buffer[11]]);

        let body = if buffer.len() > 12 {
            Some(&buffer[12..])
        } else {
            None
        };

        let retransmit_requested_from_packet_id =
            if flags & u8::from(PacketFlag::RetransmitRequest) > 0 {
                Some(u16::from_be_bytes([buffer[6], buffer[7]]))
            } else {
                None
            };

        let ack_reply = if flags & u8::from(PacketFlag::AckReply) > 0 {
            Some(u16::from_be_bytes([buffer[4], buffer[5]]))
        } else {
            None
        };

        Ok(AtemPacket {
            flags,
            session_id,
            remote_packet_id,
            body,
            retransmit_requested_from_packet_id,
            ack_reply,
        })
    }
}
